# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import unittest

from cdp.openapi_client.models.evm_user_operation import EvmUserOperation

class TestEvmUserOperation(unittest.TestCase):
    """EvmUserOperation unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional) -> EvmUserOperation:
        """Test EvmUserOperation
            include_optional is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # uncomment below to create an instance of `EvmUserOperation`
        """
        model = EvmUserOperation()
        if include_optional:
            return EvmUserOperation(
                network = 'base-sepolia',
                user_op_hash = '0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef',
                calls = [{"to":"0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48","value":"0","data":"0xa9059cbb000000000000000000000000fc807d1be4997e5c7b33e4d8d57e60c5b0f02b1a0000000000000000000000000000000000000000000000000000000000000064"},{"to":"0xdac17f958d2ee523a2206206994597c13d831ec7","value":"1000000000000000","data":"0x"}],
                status = 'pending',
                transaction_hash = '0x0123456789abcdef0123456789abcdef0123456789abcdef0123456789abcdef'
            )
        else:
            return EvmUserOperation(
                network = 'base-sepolia',
                user_op_hash = '0x1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef',
                calls = [{"to":"0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48","value":"0","data":"0xa9059cbb000000000000000000000000fc807d1be4997e5c7b33e4d8d57e60c5b0f02b1a0000000000000000000000000000000000000000000000000000000000000064"},{"to":"0xdac17f958d2ee523a2206206994597c13d831ec7","value":"1000000000000000","data":"0x"}],
                status = 'pending',
        )
        """

    def testEvmUserOperation(self):
        """Test EvmUserOperation"""
        # inst_req_only = self.make_instance(include_optional=False)
        # inst_req_and_optional = self.make_instance(include_optional=True)

if __name__ == '__main__':
    unittest.main()



from unittest.mock import AsyncMock

import pytest

from cdp.actions.solana.request_faucet import request_faucet
from cdp.openapi_client.api.faucets_api import FaucetsApi
from cdp.openapi_client.models.request_solana_faucet200_response import (
    RequestSolanaFaucet200Response as RequestSolanaFaucetResponse,
)
from cdp.openapi_client.models.request_solana_faucet_request import RequestSolanaFaucetRequest


@pytest.mark.asyncio
async def test_request_faucet_success():
    """Test successful faucet request."""
    # Setup mock faucets API
    mock_faucets_api = AsyncMock(spec=FaucetsApi)

    # Setup test data
    test_address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    test_token = "sol"
    test_tx_signature = "test_tx_signature"

    # Setup mock response
    mock_response = RequestSolanaFaucetResponse(transaction_signature=test_tx_signature)
    mock_faucets_api.request_solana_faucet = AsyncMock(return_value=mock_response)

    # Call the function
    result = await request_faucet(
        faucets=mock_faucets_api,
        address=test_address,
        token=test_token,
    )

    # Verify the API was called correctly
    mock_faucets_api.request_solana_faucet.assert_called_once_with(
        request_solana_faucet_request=RequestSolanaFaucetRequest(
            address=test_address,
            token=test_token,
        )
    )

    # Verify the response
    assert result == mock_response
    assert result.transaction_signature == test_tx_signature


@pytest.mark.asyncio
async def test_request_faucet_error():
    """Test faucet request error handling."""
    # Setup mock faucets API
    mock_faucets_api = AsyncMock(spec=FaucetsApi)

    # Setup test data
    test_address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    test_token = "sol"

    # Setup mock to raise an error
    mock_faucets_api.request_solana_faucet = AsyncMock(side_effect=Exception("API Error"))

    # Verify the error is propagated
    with pytest.raises(Exception) as exc_info:
        await request_faucet(
            faucets=mock_faucets_api,
            address=test_address,
            token=test_token,
        )

    assert str(exc_info.value) == "API Error"

    # Verify the API was called correctly
    mock_faucets_api.request_solana_faucet.assert_called_once_with(
        request_solana_faucet_request=RequestSolanaFaucetRequest(
            address=test_address,
            token=test_token,
        )
    )



from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from cdp.actions.evm.wait_for_user_operation import wait_for_user_operation
from cdp.evm_smart_account import EvmSmartAccount
from cdp.openapi_client.exceptions import ApiException
from cdp.openapi_client.models.evm_user_operation import EvmUserOperation


@pytest.mark.asyncio
@patch("cdp.actions.evm.wait_for_user_operation.time")
@patch("cdp.cdp_client.ApiClients")
async def test_wait_for_user_operation_success_immediate(mock_api_clients, mock_time):
    """Test successful completion of a user operation that is already complete."""
    mock_time.time.return_value = 1000
    mock_time.sleep = MagicMock()

    mock_smart_account = MagicMock(spec=EvmSmartAccount)
    mock_smart_account.address = "0x1234567890123456789012345678901234567890"

    mock_user_op = MagicMock(spec=EvmUserOperation)
    mock_user_op.user_op_hash = "0xuserhash123"
    mock_user_op.status = "complete"

    mock_api_clients.evm_smart_accounts.get_user_operation = AsyncMock(return_value=mock_user_op)

    result = await wait_for_user_operation(
        api_clients=mock_api_clients,
        smart_account_address=mock_smart_account.address,
        user_op_hash=mock_user_op.user_op_hash,
        timeout_seconds=20,
        interval_seconds=0.2,
    )

    assert result == mock_user_op
    mock_api_clients.evm_smart_accounts.get_user_operation.assert_called_once_with(
        mock_smart_account.address, mock_user_op.user_op_hash
    )
    mock_time.sleep.assert_not_called()


@pytest.mark.asyncio
@patch("cdp.actions.evm.wait_for_user_operation.time")
@patch("cdp.cdp_client.ApiClients")
async def test_wait_for_user_operation_success_after_poll(mock_api_clients, mock_time):
    """Test successful completion of a user operation after polling."""
    mock_time.time.side_effect = [1000, 1000.5, 1001]
    mock_time.sleep = MagicMock()

    mock_smart_account = MagicMock(spec=EvmSmartAccount)
    mock_smart_account.address = "0x1234567890123456789012345678901234567890"

    mock_initial_op = MagicMock(spec=EvmUserOperation)
    mock_initial_op.user_op_hash = "0xuserhash123"
    mock_initial_op.status = "pending"

    mock_updated_op = MagicMock(spec=EvmUserOperation)
    mock_updated_op.user_op_hash = "0xuserhash123"
    mock_updated_op.status = "complete"

    mock_api_clients.evm_smart_accounts.get_user_operation = AsyncMock(
        side_effect=[mock_initial_op, mock_updated_op]
    )

    result = await wait_for_user_operation(
        api_clients=mock_api_clients,
        smart_account_address=mock_smart_account.address,
        user_op_hash=mock_initial_op.user_op_hash,
        timeout_seconds=20,
        interval_seconds=0.2,
    )

    assert result == mock_updated_op
    assert mock_api_clients.evm_smart_accounts.get_user_operation.call_count == 2
    mock_time.sleep.assert_called_once_with(0.2)


@pytest.mark.asyncio
@patch("cdp.actions.evm.wait_for_user_operation.time")
@patch("cdp.cdp_client.ApiClients")
async def test_wait_for_user_operation_failed_status(mock_api_clients, mock_time):
    """Test handling of a user operation that completes with 'failed' status."""
    mock_time.time.side_effect = [1000, 1000.5, 1001]
    mock_time.sleep = MagicMock()

    mock_smart_account = MagicMock(spec=EvmSmartAccount)
    mock_smart_account.address = "0x1234567890123456789012345678901234567890"

    mock_initial_op = MagicMock(spec=EvmUserOperation)
    mock_initial_op.user_op_hash = "0xuserhash123"
    mock_initial_op.status = "pending"

    mock_updated_op = MagicMock(spec=EvmUserOperation)
    mock_updated_op.user_op_hash = "0xuserhash123"
    mock_updated_op.status = "failed"

    mock_api_clients.evm_smart_accounts.get_user_operation = AsyncMock(
        side_effect=[mock_initial_op, mock_updated_op]
    )

    result = await wait_for_user_operation(
        api_clients=mock_api_clients,
        smart_account_address=mock_smart_account.address,
        user_op_hash=mock_initial_op.user_op_hash,
        timeout_seconds=20,
        interval_seconds=0.2,
    )

    assert result == mock_updated_op
    assert mock_api_clients.evm_smart_accounts.get_user_operation.call_count == 2
    mock_time.sleep.assert_called_once_with(0.2)


@pytest.mark.asyncio
@patch("cdp.actions.evm.wait_for_user_operation.time")
@patch("cdp.cdp_client.ApiClients")
async def test_wait_for_user_operation_timeout(mock_api_clients, mock_time):
    """Test timeout for a user operation that never completes."""
    start_time = 1000

    # Create a sequence of times that will eventually exceed the timeout
    # First value is the initial time.time() call to set start_time
    # Then provide time values that eventually exceed start_time + timeout_seconds
    time_values = [start_time]

    # Add enough time values to handle all the API calls
    # Each iteration: check time, make API call, sleep
    current_time = start_time
    for _ in range(150):
        current_time += 0.1
        time_values.append(current_time)

    time_values[-5:] = [
        start_time + 20.1,
        start_time + 20.2,
        start_time + 20.3,
        start_time + 20.4,
        start_time + 20.5,
    ]

    mock_time.time.side_effect = time_values
    mock_time.sleep = MagicMock()

    mock_smart_account = MagicMock(spec=EvmSmartAccount)
    mock_smart_account.address = "0x1234567890123456789012345678901234567890"

    mock_pending_op = MagicMock(spec=EvmUserOperation)
    mock_pending_op.user_op_hash = "0xuserhash123"
    mock_pending_op.status = "pending"

    mock_api_clients.evm_smart_accounts.get_user_operation = AsyncMock(return_value=mock_pending_op)

    with pytest.raises(TimeoutError, match="User Operation timed out"):
        await wait_for_user_operation(
            api_clients=mock_api_clients,
            smart_account_address=mock_smart_account.address,
            user_op_hash=mock_pending_op.user_op_hash,
            timeout_seconds=20,
            interval_seconds=0.2,
        )

    assert mock_api_clients.evm_smart_accounts.get_user_operation.call_count > 1
    assert mock_time.sleep.call_count > 1


@pytest.mark.asyncio
@patch("cdp.actions.evm.wait_for_user_operation.time")
@patch("cdp.cdp_client.ApiClients")
async def test_wait_for_user_operation_api_error(mock_api_clients, mock_time):
    """Test handling of API errors during polling."""
    mock_time.time.side_effect = [1000, 1000.5]
    mock_time.sleep = MagicMock()

    mock_smart_account = MagicMock(spec=EvmSmartAccount)
    mock_smart_account.address = "0x1234567890123456789012345678901234567890"

    mock_initial_op = MagicMock(spec=EvmUserOperation)
    mock_initial_op.user_op_hash = "0xuserhash123"
    mock_initial_op.status = "pending"

    mock_api_clients.evm_smart_accounts.get_user_operation = AsyncMock(
        side_effect=ApiException(status=500, reason="Internal Server Error")
    )

    with pytest.raises(ApiException) as exc_info:
        await wait_for_user_operation(
            api_clients=mock_api_clients,
            smart_account_address=mock_smart_account.address,
            user_op_hash=mock_initial_op.user_op_hash,
            timeout_seconds=20,
            interval_seconds=0.2,
        )

    assert exc_info.value.status == 500
    assert exc_info.value.reason == "Internal Server Error"

    mock_api_clients.evm_smart_accounts.get_user_operation.assert_called_once_with(
        mock_smart_account.address, mock_initial_op.user_op_hash
    )


@pytest.mark.asyncio
@patch("cdp.actions.evm.wait_for_user_operation.time")
@patch("cdp.cdp_client.ApiClients")
async def test_wait_for_user_operation_custom_timeout_and_interval(mock_api_clients, mock_time):
    """Test using custom timeout and interval values."""
    start_time = 1000
    mock_time.time.side_effect = [
        start_time,
        start_time + 1,
        start_time + 2,
        start_time + 3,
        start_time + 11,
    ]
    mock_time.sleep = MagicMock()

    mock_smart_account = MagicMock(spec=EvmSmartAccount)
    mock_smart_account.address = "0x1234567890123456789012345678901234567890"

    mock_pending_op = MagicMock(spec=EvmUserOperation)
    mock_pending_op.user_op_hash = "0xuserhash123"
    mock_pending_op.status = "pending"

    mock_api_clients.evm_smart_accounts.get_user_operation = AsyncMock(return_value=mock_pending_op)

    with pytest.raises(TimeoutError, match="User Operation timed out"):
        await wait_for_user_operation(
            api_clients=mock_api_clients,
            smart_account_address=mock_smart_account.address,
            user_op_hash=mock_pending_op.user_op_hash,
            timeout_seconds=10,
            interval_seconds=1.0,
        )

    mock_time.sleep.assert_called_with(1.0)


@pytest.mark.asyncio
@patch("cdp.actions.evm.wait_for_user_operation.time")
@patch("cdp.cdp_client.ApiClients")
async def test_wait_for_user_operation_multiple_status_changes(mock_api_clients, mock_time):
    """Test handling of a user operation that goes through multiple status changes."""
    mock_time.time.side_effect = [1000, 1000.5, 1001, 1001.5]
    mock_time.sleep = MagicMock()

    mock_smart_account = MagicMock(spec=EvmSmartAccount)
    mock_smart_account.address = "0x1234567890123456789012345678901234567890"

    mock_initial_op = MagicMock(spec=EvmUserOperation)
    mock_initial_op.user_op_hash = "0xuserhash123"
    mock_initial_op.status = "pending"

    mock_processing_op = MagicMock(spec=EvmUserOperation)
    mock_processing_op.user_op_hash = "0xuserhash123"
    mock_processing_op.status = "processing"

    mock_complete_op = MagicMock(spec=EvmUserOperation)
    mock_complete_op.user_op_hash = "0xuserhash123"
    mock_complete_op.status = "complete"

    mock_api_clients.evm_smart_accounts.get_user_operation = AsyncMock(
        side_effect=[mock_initial_op, mock_processing_op, mock_complete_op]
    )

    result = await wait_for_user_operation(
        api_clients=mock_api_clients,
        smart_account_address=mock_smart_account.address,
        user_op_hash=mock_initial_op.user_op_hash,
        timeout_seconds=20,
        interval_seconds=0.2,
    )

    assert result == mock_complete_op
    assert mock_api_clients.evm_smart_accounts.get_user_operation.call_count == 3
    assert mock_time.sleep.call_count == 2


@pytest.mark.asyncio
@patch("cdp.actions.evm.wait_for_user_operation.time")
@patch("cdp.cdp_client.ApiClients")
async def test_wait_for_user_operation_invalid_user_op_hash(mock_api_clients, mock_time):
    """Test handling of an API error when user_op_hash is invalid."""
    mock_time.time.return_value = 1000
    mock_time.sleep = MagicMock()

    mock_smart_account = MagicMock(spec=EvmSmartAccount)
    mock_smart_account.address = "0x1234567890123456789012345678901234567890"

    mock_op = MagicMock(spec=EvmUserOperation)
    mock_op.user_op_hash = "invalid-hash"
    mock_op.status = "pending"

    mock_api_clients.evm_smart_accounts.get_user_operation = AsyncMock(
        side_effect=ApiException(status=404, reason="User operation not found")
    )

    with pytest.raises(ApiException) as exc_info:
        await wait_for_user_operation(
            api_clients=mock_api_clients,
            smart_account_address=mock_smart_account.address,
            user_op_hash=mock_op.user_op_hash,
            timeout_seconds=20,
            interval_seconds=0.2,
        )

    assert exc_info.value.status == 404
    assert exc_info.value.reason == "User operation not found"



import base64

import pytest
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed25519


@pytest.fixture
def ec_private_key_factory():
    """Create and return a factory for EC private keys.

    Returns:
        callable: A factory function that creates PEM-encoded EC private keys

    """

    def _create_key():
        private_key = ec.generate_private_key(ec.SECP256K1())
        pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        )
        return pem.decode()

    return _create_key


@pytest.fixture
def ed25519_private_key_factory():
    """Create and return a factory for Ed25519 private keys.

    Returns:
        callable: A factory function that creates base64-encoded Ed25519 private keys

    """

    def _create_key():
        private_key = ed25519.Ed25519PrivateKey.generate()
        private_bytes = private_key.private_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PrivateFormat.Raw,
            encryption_algorithm=serialization.NoEncryption(),
        )
        public_bytes = private_key.public_key().public_bytes(
            encoding=serialization.Encoding.Raw,
            format=serialization.PublicFormat.Raw,
        )
        return base64.b64encode(private_bytes + public_bytes).decode()

    return _create_key



import pytest

from cdp.auth.utils.jwt import JwtOptions


@pytest.fixture
def jwt_options_factory():
    """Create and return a factory for JwtOptions fixtures.

    Returns:
        callable: A factory function that creates JwtOptions instances

    """

    def _create_options(
        api_key_id="test-key-id",
        api_key_secret="dummy-secret",
        request_method="GET",
        request_host="https://api.cdp.coinbase.com",
        request_path="/v1/test",
        expires_in=120,
    ):
        return JwtOptions(
            api_key_id=api_key_id,
            api_key_secret=api_key_secret,
            request_method=request_method,
            request_host=request_host,
            request_path=request_path,
            expires_in=expires_in,
        )

    return _create_options


@pytest.fixture
def websocket_jwt_options_factory():
    """Create and return a factory for WebSocket JwtOptions fixtures with null request parameters.

    Returns:
        callable: A factory function that creates JwtOptions instances for WebSocket

    """

    def _create_options(
        api_key_id="test-key-id",
        api_key_secret="dummy-secret",
        expires_in=120,
    ):
        return JwtOptions(
            api_key_id=api_key_id,
            api_key_secret=api_key_secret,
            request_method=None,
            request_host=None,
            request_path=None,
            expires_in=expires_in,
        )

    return _create_options



from pydantic import BaseModel


class FundOperationResult(BaseModel):
    """The result of a fund operation."""

    """
    The result of a fund operation.

    Attributes:
        id: The transfer that was created to fund the account.
        network: The network that the transfer was created on.
        target_amount: The target amount that will be received.
        target_currency: The currency that will be received.
        status: The status of the fund operation.
        transaction_hash: The transaction hash of the transfer. This is None if the transfer is pending.
    """

    id: str
    network: str
    target_amount: str
    target_currency: str
    status: str
    transaction_hash: str | None



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, StrictStr, field_validator
from typing import Any, ClassVar, Dict, List, Optional
from typing_extensions import Annotated
from cdp.openapi_client.models.fee import Fee
from cdp.openapi_client.models.transfer_source import TransferSource
from cdp.openapi_client.models.transfer_target import TransferTarget
from typing import Optional, Set
from typing_extensions import Self

class Transfer(BaseModel):
    """
    The transfer object.
    """ # noqa: E501
    id: Annotated[str, Field(strict=True)] = Field(description="The ID of the transfer.")
    source_type: StrictStr = Field(description="The type of the source of the transfer.", alias="sourceType")
    source: TransferSource
    target_type: StrictStr = Field(description="The type of the target of the transfer.", alias="targetType")
    target: TransferTarget
    source_amount: StrictStr = Field(description="The amount the source will transfer.", alias="sourceAmount")
    source_currency: StrictStr = Field(description="The currency the source will transfer.", alias="sourceCurrency")
    target_amount: StrictStr = Field(description="The amount the target will receive.", alias="targetAmount")
    target_currency: StrictStr = Field(description="The currency the target will receive.", alias="targetCurrency")
    user_amount: StrictStr = Field(description="The amount the customer put in to transfer.", alias="userAmount")
    user_currency: StrictStr = Field(description="The currency the customer put in to transfer.", alias="userCurrency")
    fees: List[Fee] = Field(description="The fees for the transfer.")
    status: StrictStr = Field(description="The status of the transfer.")
    created_at: StrictStr = Field(description="The UTC date and time in ISO 8601 format the transfer was created.", alias="createdAt")
    updated_at: StrictStr = Field(description="The UTC date and time in ISO 8601 format the transfer was updated.", alias="updatedAt")
    transaction_hash: Optional[StrictStr] = Field(default=None, description="The transaction hash or transaction signature of the transfer.", alias="transactionHash")
    __properties: ClassVar[List[str]] = ["id", "sourceType", "source", "targetType", "target", "sourceAmount", "sourceCurrency", "targetAmount", "targetCurrency", "userAmount", "userCurrency", "fees", "status", "createdAt", "updatedAt", "transactionHash"]

    @field_validator('id')
    def id_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if not re.match(r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$", value):
            raise ValueError(r"must validate the regular expression /^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$/")
        return value

    @field_validator('source_type')
    def source_type_validate_enum(cls, value):
        """Validates the enum"""
        if value not in set(['payment_method']):
            raise ValueError("must be one of enum values ('payment_method')")
        return value

    @field_validator('target_type')
    def target_type_validate_enum(cls, value):
        """Validates the enum"""
        if value not in set(['crypto_rail']):
            raise ValueError("must be one of enum values ('crypto_rail')")
        return value

    @field_validator('status')
    def status_validate_enum(cls, value):
        """Validates the enum"""
        if value not in set(['created', 'pending', 'started', 'completed', 'failed']):
            raise ValueError("must be one of enum values ('created', 'pending', 'started', 'completed', 'failed')")
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of Transfer from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of source
        if self.source:
            _dict['source'] = self.source.to_dict()
        # override the default output from pydantic by calling `to_dict()` of target
        if self.target:
            _dict['target'] = self.target.to_dict()
        # override the default output from pydantic by calling `to_dict()` of each item in fees (list)
        _items = []
        if self.fees:
            for _item_fees in self.fees:
                if _item_fees:
                    _items.append(_item_fees.to_dict())
            _dict['fees'] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of Transfer from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "id": obj.get("id"),
            "sourceType": obj.get("sourceType"),
            "source": TransferSource.from_dict(obj["source"]) if obj.get("source") is not None else None,
            "targetType": obj.get("targetType"),
            "target": TransferTarget.from_dict(obj["target"]) if obj.get("target") is not None else None,
            "sourceAmount": obj.get("sourceAmount"),
            "sourceCurrency": obj.get("sourceCurrency"),
            "targetAmount": obj.get("targetAmount"),
            "targetCurrency": obj.get("targetCurrency"),
            "userAmount": obj.get("userAmount"),
            "userCurrency": obj.get("userCurrency"),
            "fees": [Fee.from_dict(_item) for _item in obj["fees"]] if obj.get("fees") is not None else None,
            "status": obj.get("status"),
            "createdAt": obj.get("createdAt"),
            "updatedAt": obj.get("updatedAt"),
            "transactionHash": obj.get("transactionHash")
        })
        return _obj





import base64
import random
import time
import uuid
from datetime import datetime
from typing import Any
from urllib.parse import urlparse

import jwt
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec, ed25519
from pydantic import BaseModel, Field, field_validator


class JwtOptions(BaseModel):
    r"""Configuration options for JWT generation.

    This class holds all necessary parameters for generating a JWT token
    for authenticating with Coinbase's REST APIs. It supports both EC (ES256)
    and Ed25519 (EdDSA) keys.

    Attributes:
        api_key_id - The API key ID
            Examples:
                'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
                'organizations/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/apiKeys/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
        api_key_secret - The API key secret
            Examples:
                'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx==' (Edwards key (Ed25519))
                '-----BEGIN EC PRIVATE KEY-----\\n...\\n...\\n...==\\n-----END EC PRIVATE KEY-----\\n' (EC key (ES256))
        request_method - The HTTP method for the request (e.g. 'GET', 'POST'), or None for JWTs intended for websocket connections
        request_host - The host for the request (e.g. 'api.cdp.coinbase.com'), or None for JWTs intended for websocket connections
        request_path - The path for the request (e.g. '/platform/v1/wallets'), or None for JWTs intended for websocket connections
        expires_in - Optional expiration time in seconds (defaults to 120)
        audience - Optional audience claim for the JWT

    """

    api_key_id: str = Field(..., description="The API key ID")
    api_key_secret: str = Field(..., description="The API key secret")
    request_method: str | None = Field(
        None,
        description="The HTTP method for the request or None for JWTs intended for websocket connections",
    )
    request_host: str | None = Field(
        None,
        description="The host for the request or None for JWTs intended for websocket connections",
    )
    request_path: str | None = Field(
        None,
        description="The path for the request or None for JWTs intended for websocket connections",
    )
    expires_in: int | None = Field(120, description="Optional expiration time in seconds")
    audience: list[str] | None = Field(None, description="Optional audience claim for the JWT")

    @field_validator("request_method")
    @classmethod
    def validate_request_method(cls, v: str | None) -> str | None:
        """Validate the HTTP request method.

        Args:
            v: The request method to validate, or None for JWTs intended for websocket connections

        Returns:
            The validated request method in uppercase, or None

        Raises:
            ValueError: If the request method is invalid

        """
        if v is None:
            return None

        valid_methods = ["GET", "POST", "PUT", "DELETE", "PATCH"]
        upper_method = v.upper()
        if upper_method not in valid_methods:
            raise ValueError(f"Invalid request method. Must be one of: {', '.join(valid_methods)}")
        return upper_method


class WalletJwtOptions(BaseModel):
    """Configuration options for Wallet Auth JWT generation.

    This class holds all necessary parameters for generating a Wallet Auth JWT
    for authenticating with endpoints that require wallet authentication.

    Attributes:
        wallet_auth_key - The wallet authentication key
        request_method - The HTTP method for the request (e.g. 'GET', 'POST')
        request_host - The host for the request (e.g. 'api.cdp.coinbase.com')
        request_path - The path for the request (e.g. '/platform/v1/wallets/{wallet_id}/addresses')
        request_data - The request data for the request (e.g. { "wallet_id": "1234567890" })

    """

    wallet_auth_key: str = Field(..., description="The wallet authentication key")
    request_method: str = Field(..., description="The HTTP method for the request")
    request_host: str = Field(..., description="The host for the request")
    request_path: str = Field(..., description="The path for the request")
    request_data: dict[str, Any] = Field(..., description="The request data")

    @field_validator("request_method")
    @classmethod
    def validate_request_method(cls, v: str) -> str:
        """Validate the HTTP request method.

        Args:
            v: The request method to validate

        Returns:
            The validated request method in uppercase

        Raises:
            ValueError: If the request method is invalid

        """
        valid_methods = ["GET", "POST", "PUT", "DELETE", "PATCH"]
        upper_method = v.upper()
        if upper_method not in valid_methods:
            raise ValueError(f"Invalid request method. Must be one of: {', '.join(valid_methods)}")
        return upper_method


def generate_jwt(options: JwtOptions) -> str:
    """Generate a JWT (Bearer token) for authenticating with Coinbase's APIs.

    Supports both EC (ES256) and Ed25519 (EdDSA) keys. Also supports JWTs meant for
    websocket connections by allowing request_method, request_host, and request_path
    to all be None, in which case the 'uris' claim is omitted from the JWT.

    Args:
        options: The configuration options for generating the JWT

    Returns:
        The generated JWT (Bearer token) string

    Raises:
        ValueError: If required parameters are missing, invalid, or if JWT signing fails

    """
    # Validate required parameters
    if not options.api_key_id:
        raise ValueError("Key ID is required")
    if not options.api_key_secret:
        raise ValueError("Private key is required")

    # Check if we have a REST API request or a websocket connection
    has_all_uri_params = all([options.request_method, options.request_host, options.request_path])
    has_no_uri_params = all(
        param is None
        for param in [options.request_method, options.request_host, options.request_path]
    )

    # Ensure we either have all request parameters or none (for websocket)
    if not (has_all_uri_params or has_no_uri_params):
        raise ValueError(
            "Either all request details (method, host, path) must be provided, or all must be None for JWTs intended for websocket connections"
        )

    try:
        # Parse the private key
        private_key = _parse_private_key(options.api_key_secret)

        # Determine algorithm based on key type
        if isinstance(private_key, ec.EllipticCurvePrivateKey):
            algorithm = "ES256"
        elif isinstance(private_key, ed25519.Ed25519PrivateKey):
            algorithm = "EdDSA"
        else:
            raise ValueError("Unsupported key type")

        # Create header with nonce
        header = {
            "alg": algorithm,
            "kid": options.api_key_id,
            "typ": "JWT",
            "nonce": _generate_nonce(),
        }

        # Create claims with timing
        now = int(time.time())
        expires_in = options.expires_in or 120  # Default to 120 seconds

        claims = {
            "sub": options.api_key_id,
            "iss": "cdp",
            "aud": options.audience if options.audience is not None else ["cdp_service"],
            "nbf": now,
            "exp": now + expires_in,
        }

        # Add the uris claim only for JWTs intended for REST API requests, not for websocket connections
        if has_all_uri_params:
            # Build the URI claim
            parsed_url = urlparse(f"{options.request_host}{options.request_path}")
            uri = f"{options.request_method} {parsed_url.netloc}{parsed_url.path}"
            claims["uris"] = [uri]

        # Generate the JWT
        return jwt.encode(claims, private_key, algorithm=algorithm, headers=header)

    except Exception as error:
        raise ValueError(f"Failed to generate JWT: {error!s}") from error


def generate_wallet_jwt(options: WalletJwtOptions) -> str:
    """Build a wallet authentication JWT for the given API endpoint URL.

    Used for authenticating with specific endpoints that require wallet authentication.

    Args:
        options: The configuration options for generating the wallet auth JWT

    Returns:
        The generated JWT (Bearer token) string

    Raises:
        ValueError: If required parameters are missing or if JWT signing fails

    """
    if not options.wallet_auth_key:
        raise ValueError("Server Wallet Secret is not defined")

    uri = f"{options.request_method} {options.request_host}{options.request_path}"
    now = int(datetime.now().timestamp())

    claims = {"uris": [uri], "iat": now, "nbf": now, "jti": str(uuid.uuid4())}

    if options.request_data:
        claims["req"] = options.request_data

    try:
        der_bytes = serialization.load_der_private_key(
            base64.b64decode(options.wallet_auth_key), password=None
        )

        # Generate and sign the token
        token = jwt.encode(
            claims,
            der_bytes,  # Use the private key directly
            algorithm="ES256",
            headers={"typ": "JWT"},
        )

        return token

    except Exception as error:
        raise ValueError(f"Could not create the EC key: {error!s}") from error


def _parse_private_key(
    key_data: str,
) -> ec.EllipticCurvePrivateKey | ed25519.Ed25519PrivateKey:
    """Parse a private key from PEM or base64 format.

    Args:
        key_data: The private key data in either PEM (EC) or base64 (Ed25519) format

    Returns:
        The parsed private key object

    Raises:
        ValueError: If the key cannot be parsed or is of an unsupported type

    """
    # First try parsing as PEM (EC key)
    try:
        key = serialization.load_pem_private_key(key_data.encode(), password=None)
        if isinstance(key, ec.EllipticCurvePrivateKey):
            return key
    except Exception:
        pass

    # Try parsing as base64 (Ed25519 key)
    try:
        decoded = base64.b64decode(key_data)
        if len(decoded) == 64:
            # Ed25519 keys are 64 bytes - first 32 is seed, second 32 is public key
            seed = decoded[:32]
            return ed25519.Ed25519PrivateKey.from_private_bytes(seed)
    except Exception:
        pass

    raise ValueError("Key must be either PEM EC key or base64 Ed25519 key")


def _generate_nonce() -> str:
    """Generate a random nonce for the JWT header.

    Returns:
        A 16-character random string of digits

    """
    return "".join(random.choices("0123456789", k=16))



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import unittest

from cdp.openapi_client.models.get_swap_price_response_wrapper import GetSwapPriceResponseWrapper

class TestGetSwapPriceResponseWrapper(unittest.TestCase):
    """GetSwapPriceResponseWrapper unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional) -> GetSwapPriceResponseWrapper:
        """Test GetSwapPriceResponseWrapper
            include_optional is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # uncomment below to create an instance of `GetSwapPriceResponseWrapper`
        """
        model = GetSwapPriceResponseWrapper()
        if include_optional:
            return GetSwapPriceResponseWrapper(
                block_number = '17038723',
                to_amount = '1000000000000000000',
                to_token = '0x7F5c764cBc14f9669B88837ca1490cCa17c31607',
                fees = {"gasFee":{"amount":"1000000000000000000","token":"0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"},"protocolFee":{"amount":"1000000000000000000","token":"0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"}},
                issues = {"allowance":{"currentAllowance":"1000000000","spender":"0x000000000022D473030F116dDEE9F6B43aC78BA3"},"balance":{"token":"0x6B175474E89094C44Da98b954EedeAC495271d0F","currentBalance":"900000000000000000","requiredBalance":"1000000000000000000"},"simulationIncomplete":false},
                liquidity_available = False,
                min_to_amount = '900000000000000000',
                from_amount = '1000000000000000000',
                from_token = '0x6B175474E89094C44Da98b954EedeAC495271d0F',
                gas = '100000',
                gas_price = '1000000000'
            )
        else:
            return GetSwapPriceResponseWrapper(
                block_number = '17038723',
                to_amount = '1000000000000000000',
                to_token = '0x7F5c764cBc14f9669B88837ca1490cCa17c31607',
                fees = {"gasFee":{"amount":"1000000000000000000","token":"0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"},"protocolFee":{"amount":"1000000000000000000","token":"0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"}},
                issues = {"allowance":{"currentAllowance":"1000000000","spender":"0x000000000022D473030F116dDEE9F6B43aC78BA3"},"balance":{"token":"0x6B175474E89094C44Da98b954EedeAC495271d0F","currentBalance":"900000000000000000","requiredBalance":"1000000000000000000"},"simulationIncomplete":false},
                liquidity_available = False,
                min_to_amount = '900000000000000000',
                from_amount = '1000000000000000000',
                from_token = '0x6B175474E89094C44Da98b954EedeAC495271d0F',
                gas = '100000',
                gas_price = '1000000000',
        )
        """

    def testGetSwapPriceResponseWrapper(self):
        """Test GetSwapPriceResponseWrapper"""
        # inst_req_only = self.make_instance(include_optional=False)
        # inst_req_and_optional = self.make_instance(include_optional=True)

if __name__ == '__main__':
    unittest.main()



"""Main swap functionality."""

from eth_account.signers.base import BaseAccount

from cdp.actions.evm.swap.types import SwapOptions, SwapResult, SwapStrategy
from cdp.api_clients import ApiClients


async def swap(
    api_clients: ApiClients,
    from_account: BaseAccount,
    swap_options: SwapOptions,
    swap_strategy: SwapStrategy,
) -> SwapResult:
    """Execute a token swap.

    This function supports two patterns:
    1. Direct swap parameters in SwapOptions
    2. Pre-created swap quote in SwapOptions

    Args:
        api_clients: The API clients instance
        from_account: The account to swap from
        swap_options: The swap options containing either direct parameters or a swap quote
        swap_strategy: The strategy to use for executing the swap

    Returns:
        SwapResult: The result of the swap transaction

    Raises:
        ValueError: If the swap options are invalid
        Exception: If the swap fails

    Examples:
        **Using direct parameters**:
        ```python
        result = await swap(
            api_clients=api_clients,
            from_account=account,
            swap_options=SwapOptions(
                network="base",
                from_token="0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",  # USDC
                to_token="0x4200000000000000000000000000000000000006",  # WETH
                from_amount="1000000",  # 1 USDC
                slippage_bps=100
            ),
            swap_strategy=AccountSwapStrategy()
        )
        ```

        **Using pre-created swap quote**:
        ```python
        # First create the swap
        swap_quote = await cdp.evm.create_swap_quote(...)

        # Then execute it
        result = await swap(
            api_clients=api_clients,
            from_account=account,
            swap_options=SwapOptions(swap_quote=swap_quote),
            swap_strategy=AccountSwapStrategy()
        )
        ```

    """
    # Import EVM client here to avoid circular imports
    from cdp.evm_client import EvmClient

    # Create an EVM client instance
    evm_client = EvmClient(api_clients)

    # Determine which pattern is being used and get swap data
    swap_data = None
    network = None
    permit2_signature = None

    if swap_options.swap_quote is not None:
        # Pattern 1: Pre-created swap quote
        swap_quote = swap_options.swap_quote

        # Handle Permit2 signature if required
        if swap_quote.requires_signature and swap_quote.permit2_data:
            # Sign the Permit2 typed data
            typed_data = swap_quote.permit2_data.eip712
            permit2_signature = await from_account.sign_typed_data(
                domain=typed_data["domain"],
                types=typed_data["types"],
                primary_type=typed_data["primaryType"],
                message=typed_data["message"],
            )

        swap_data = swap_quote
        network = swap_quote.network

    else:
        # Pattern 2: Direct parameters
        # Create the swap quote from parameters
        swap_quote = await evm_client.create_swap_quote(
            from_token=swap_options.from_token,
            to_token=swap_options.to_token,
            from_amount=swap_options.from_amount,
            network=swap_options.network,
            taker=from_account.address,  # Always use account address as taker
            slippage_bps=swap_options.slippage_bps,
        )

        # Handle Permit2 signature if required
        if swap_quote.requires_signature and swap_quote.permit2_data:
            # Sign the Permit2 typed data
            typed_data = swap_quote.permit2_data.eip712
            permit2_signature = await from_account.sign_typed_data(
                domain=typed_data["domain"],
                types=typed_data["types"],
                primary_type=typed_data["primaryType"],
                message=typed_data["message"],
            )

        swap_data = swap_quote
        network = swap_options.network

    # Execute the swap using the appropriate strategy
    result = await swap_strategy.execute_swap(
        api_clients=api_clients,
        from_account=from_account,
        swap_data=swap_data,
        network=network,
        permit2_signature=permit2_signature,
    )

    return result



from unittest.mock import AsyncMock, MagicMock

import pytest

from cdp.openapi_client.models.request_solana_faucet200_response import (
    RequestSolanaFaucet200Response as RequestSolanaFaucetResponse,
)
from cdp.openapi_client.models.request_solana_faucet_request import RequestSolanaFaucetRequest
from cdp.openapi_client.models.sign_solana_message200_response import (
    SignSolanaMessage200Response as SignSolanaMessageResponse,
)
from cdp.openapi_client.models.sign_solana_message_request import SignSolanaMessageRequest
from cdp.openapi_client.models.sign_solana_transaction200_response import (
    SignSolanaTransaction200Response as SignSolanaTransactionResponse,
)
from cdp.openapi_client.models.sign_solana_transaction_request import SignSolanaTransactionRequest
from cdp.openapi_client.models.solana_account import SolanaAccount as SolanaAccountModel
from cdp.solana_account import SolanaAccount


def test_initialization():
    """Test that the SolanaAccount initializes correctly."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    name = "test-account"

    solana_account_model = SolanaAccountModel(address=address, name=name)
    mock_api_clients = MagicMock()
    account = SolanaAccount(solana_account_model, mock_api_clients)

    assert account.address == address
    assert account.name == name


def test_str_representation():
    """Test the string representation of SolanaAccount."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    solana_account_model = SolanaAccountModel(address=address)
    mock_api_clients = MagicMock()
    account = SolanaAccount(solana_account_model, mock_api_clients)

    expected = f"Solana Account Address: {address}"
    assert str(account) == expected


def test_repr_representation():
    """Test the repr representation of SolanaAccount."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    solana_account_model = SolanaAccountModel(address=address)
    mock_api_clients = MagicMock()
    account = SolanaAccount(solana_account_model, mock_api_clients)

    expected = f"Solana Account Address: {address}"
    assert repr(account) == expected


@pytest.mark.asyncio
async def test_request_faucet():
    """Test request_faucet method."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    name = "test-account"
    solana_account_model = SolanaAccountModel(address=address, name=name)

    mock_faucets_api = AsyncMock()
    mock_api_clients = MagicMock()
    mock_api_clients.faucets = mock_faucets_api

    mock_response = RequestSolanaFaucetResponse(transaction_signature="test_tx_signature")
    mock_faucets_api.request_solana_faucet = AsyncMock(return_value=mock_response)

    account = SolanaAccount(solana_account_model, mock_api_clients)

    result = await account.request_faucet(token="sol")

    mock_faucets_api.request_solana_faucet.assert_called_once_with(
        request_solana_faucet_request=RequestSolanaFaucetRequest(
            address=address,
            token="sol",
        )
    )
    assert result == mock_response
    assert result.transaction_signature == "test_tx_signature"


@pytest.mark.asyncio
async def test_sign_message():
    """Test sign_message method."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    name = "test-account"
    solana_account_model = SolanaAccountModel(address=address, name=name)

    mock_solana_accounts_api = AsyncMock()
    mock_api_clients = MagicMock()
    mock_api_clients.solana_accounts = mock_solana_accounts_api

    mock_response = SignSolanaMessageResponse(signature="test_signature")
    mock_solana_accounts_api.sign_solana_message = AsyncMock(return_value=mock_response)

    account = SolanaAccount(solana_account_model, mock_api_clients)

    test_message = "Hello, Solana!"
    test_idempotency_key = "test-idempotency-key"
    result = await account.sign_message(
        message=test_message,
        idempotency_key=test_idempotency_key,
    )

    mock_solana_accounts_api.sign_solana_message.assert_called_once_with(
        sign_solana_message_request=SignSolanaMessageRequest(
            message=test_message,
        ),
        address=address,
        x_idempotency_key=test_idempotency_key,
    )
    assert result == mock_response
    assert result.signature == "test_signature"


@pytest.mark.asyncio
async def test_sign_transaction():
    """Test sign_transaction method."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    name = "test-account"
    solana_account_model = SolanaAccountModel(address=address, name=name)

    mock_solana_accounts_api = AsyncMock()
    mock_api_clients = MagicMock()
    mock_api_clients.solana_accounts = mock_solana_accounts_api

    mock_response = SignSolanaTransactionResponse(signed_transaction="test_signed_transaction")
    mock_solana_accounts_api.sign_solana_transaction = AsyncMock(return_value=mock_response)

    account = SolanaAccount(solana_account_model, mock_api_clients)

    test_transaction = "test_transaction_data"
    test_idempotency_key = "test-idempotency-key"
    result = await account.sign_transaction(
        transaction=test_transaction,
        idempotency_key=test_idempotency_key,
    )

    mock_solana_accounts_api.sign_solana_transaction.assert_called_once_with(
        sign_solana_transaction_request=SignSolanaTransactionRequest(
            transaction=test_transaction,
        ),
        address=address,
        x_idempotency_key=test_idempotency_key,
    )
    assert result == mock_response
    assert result.signed_transaction == "test_signed_transaction"


@pytest.mark.asyncio
async def test_request_faucet_error():
    """Test request_faucet error handling."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    name = "test-account"
    solana_account_model = SolanaAccountModel(address=address, name=name)

    mock_faucets_api = AsyncMock()
    mock_api_clients = MagicMock()
    mock_api_clients.faucets = mock_faucets_api

    mock_faucets_api.request_solana_faucet = AsyncMock(side_effect=Exception("API Error"))

    account = SolanaAccount(solana_account_model, mock_api_clients)

    with pytest.raises(Exception) as exc_info:
        await account.request_faucet(token="sol")

    assert str(exc_info.value) == "API Error"


@pytest.mark.asyncio
async def test_sign_message_error():
    """Test sign_message error handling."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    name = "test-account"
    solana_account_model = SolanaAccountModel(address=address, name=name)

    mock_solana_accounts_api = AsyncMock()
    mock_api_clients = MagicMock()
    mock_api_clients.solana_accounts = mock_solana_accounts_api

    mock_solana_accounts_api.sign_solana_message = AsyncMock(side_effect=Exception("API Error"))

    account = SolanaAccount(solana_account_model, mock_api_clients)

    with pytest.raises(Exception) as exc_info:
        await account.sign_message(message="test_message")

    assert str(exc_info.value) == "API Error"


@pytest.mark.asyncio
async def test_sign_transaction_error():
    """Test sign_transaction error handling."""
    address = "14grJpemFaf88c8tiVb77W7TYg2W3ir6pfkKz3YjhhZ5"
    name = "test-account"
    solana_account_model = SolanaAccountModel(address=address, name=name)

    mock_solana_accounts_api = AsyncMock()
    mock_api_clients = MagicMock()
    mock_api_clients.solana_accounts = mock_solana_accounts_api

    mock_solana_accounts_api.sign_solana_transaction = AsyncMock(side_effect=Exception("API Error"))

    account = SolanaAccount(solana_account_model, mock_api_clients)

    with pytest.raises(Exception) as exc_info:
        await account.sign_transaction(transaction="test_transaction")

    assert str(exc_info.value) == "API Error"



import json
from urllib.parse import urlparse

from urllib3.util import Retry

from cdp import __version__
from cdp.auth.utils.http import GetAuthHeadersOptions, get_auth_headers
from cdp.openapi_client import rest
from cdp.openapi_client.api_client import ApiClient
from cdp.openapi_client.api_response import ApiResponse, T as ApiResponseT
from cdp.openapi_client.configuration import Configuration
from cdp.openapi_client.constants import ERROR_DOCS_PAGE_URL, SDK_DEFAULT_SOURCE
from cdp.openapi_client.errors import ApiError, is_openapi_error
from cdp.openapi_client.exceptions import ApiException


class CdpApiClient(ApiClient):
    """CDP API Client that handles authentication and API calls for Coinbase."""

    def __init__(
        self,
        api_key_id: str,
        api_key_secret: str,
        wallet_secret: str = None,
        debugging: bool = False,
        base_path: str = "https://api.cdp.coinbase.com/platform",
        max_network_retries: int = 3,
        source: str = SDK_DEFAULT_SOURCE,
        source_version: str = __version__,
    ):
        """Initialize the CDP API Client.

        Args:
            api_key_id (str): The API key id for authentication.
            api_key_secret (str): The API key secret for authentication.
            wallet_secret (str): The wallet secret for authentication.
            debugging (bool): Whether debugging is enabled.
            base_path (str, optional): The base URL for the API. Defaults to "https://api.cdp.coinbase.com/platform".
            max_network_retries (int): The maximum number of network retries. Defaults to 3.
            source (str): Specifies whether the sdk is being used directly or if it's an Agentkit extension.
            source_version (str): The version of the source package.

        """
        retry_strategy = self._get_retry_strategy(max_network_retries)
        configuration = Configuration(host=base_path, retries=retry_strategy)
        super().__init__(configuration)

        self.api_key_id = api_key_id
        self.api_key_secret = api_key_secret
        self.wallet_secret = wallet_secret
        self.source = source
        self.source_version = source_version
        self._debugging = debugging

    async def call_api(
        self,
        method,
        url,
        header_params=None,
        body=None,
        post_params=None,
        _request_timeout=None,
    ) -> rest.RESTResponse:
        """Make the HTTP request (asynchronous)."""
        if self._debugging is True:
            print(f"CDP API REQUEST: {method} {url}")

        # Parse URL for auth headers
        parsed_url = urlparse(
            url if url.startswith("http") else self.configuration.host + url
        )

        # Get auth headers
        auth_headers = get_auth_headers(
            GetAuthHeadersOptions(
                api_key_id=self.api_key_id,
                api_key_secret=self.api_key_secret,
                request_method=method,
                request_host=parsed_url.netloc,
                request_path=parsed_url.path,
                request_body=body,
                wallet_secret=self.wallet_secret,
                source=self.source,
                source_version=self.source_version,
            )
        )

        # Merge headers
        request_headers = header_params or {}
        request_headers.update(auth_headers)

        if self._debugging is True:
            print(f"Request headers: {request_headers}")

        # Make request through parent class
        try:
            response = await super().call_api(
                method, url, request_headers, body, post_params, _request_timeout
            )
            return response
        except ApiException as e:
            if self._debugging:
                print(f"Error: {e}")

            # Try to parse response body as JSON
            try:
                error_data = json.loads(e.body) if e.body else None
                if error_data and is_openapi_error(error_data):
                    # Handle OpenAPI formatted error
                    raise ApiError(
                        http_code=e.status,
                        error_type=error_data["errorType"],
                        error_message=error_data["errorMessage"],
                        correlation_id=error_data.get("correlationId"),
                        error_link=error_data.get(
                            "errorLink",
                            f"{ERROR_DOCS_PAGE_URL}#{error_data['errorType'].lower()}",
                        ),
                    ) from None
            except (json.JSONDecodeError, AttributeError):
                pass

            # Handle HTTP status code based errors
            if e.status == 401:
                raise ApiError(
                    http_code=401,
                    error_type="unauthorized",
                    error_message="Unauthorized.",
                    error_link=f"{ERROR_DOCS_PAGE_URL}#unauthorized",
                ) from None
            elif e.status == 404:
                raise ApiError(
                    http_code=404,
                    error_type="not_found",
                    error_message="API not found",
                    error_link=f"{ERROR_DOCS_PAGE_URL}#not_found",
                ) from None
            elif e.status == 502:
                raise ApiError(
                    http_code=502,
                    error_type="bad_gateway",
                    error_message="Bad gateway.",
                    error_link=ERROR_DOCS_PAGE_URL,
                ) from None
            elif e.status == 503:
                raise ApiError(
                    http_code=503,
                    error_type="service_unavailable",
                    error_message="Service unavailable. Please try again later.",
                    error_link=ERROR_DOCS_PAGE_URL,
                ) from None

            # Default to unexpected error
            raise ApiError(
                http_code=e.status or 500,
                error_type="unexpected_error",
                error_message="An unexpected error occurred.",
                error_link=ERROR_DOCS_PAGE_URL,
            ) from None
        except Exception as e:
            if self._debugging:
                print(f"Error: {e}")

            # Handle network errors
            if "Connection refused" in str(e):
                raise ApiError(
                    http_code=503,
                    error_type="service_unavailable",
                    error_message="Network error, unable to reach the service.",
                    error_link=ERROR_DOCS_PAGE_URL,
                ) from None

            # Default unexpected error
            raise ApiError(
                http_code=500,
                error_type="unexpected_error",
                error_message=f"An unexpected error occurred: {e!s}",
                error_link=ERROR_DOCS_PAGE_URL,
            ) from None

    def response_deserialize(
        self,
        response_data: rest.RESTResponse,
        response_types_map: dict[str, ApiResponseT] | None = None,
    ) -> ApiResponse[ApiResponseT]:
        """Deserialize the API response.

        Args:
            response_data: REST response data.
            response_types_map: Map of response types.

        Returns:
            ApiResponse[ApiResponseT]

        """
        try:
            return super().response_deserialize(response_data, response_types_map)
        except ApiException as e:
            # Try to parse response body as JSON
            try:
                error_data = json.loads(e.body) if e.body else None
                if error_data and is_openapi_error(error_data):
                    raise ApiError(
                        http_code=e.status,
                        error_type=error_data["errorType"],
                        error_message=error_data["errorMessage"],
                        correlation_id=error_data.get("correlationId"),
                        error_link=error_data.get(
                            "errorLink",
                            f"{ERROR_DOCS_PAGE_URL}#{error_data['errorType'].lower()}",
                        ),
                    ) from None
            except (json.JSONDecodeError, AttributeError) as parse_error:
                # If we can't parse as OpenAPI error, include the parse error details
                raise ApiError(
                    http_code=e.status,
                    error_type="unexpected_error",
                    error_message=f"An unexpected error occurred: {parse_error!s}. Original error message: {e!s}.",
                    error_link=f"{ERROR_DOCS_PAGE_URL}",
                ) from None

    def _get_retry_strategy(self, max_network_retries: int) -> Retry:
        """Return the retry strategy for the CDP API Client.

        Args:
            max_network_retries (int): The maximum number of network retries.

        Returns:
            Retry: The retry strategy.

        """
        return Retry(
            total=max_network_retries,  # Number of total retries
            status_forcelist=[500, 502, 503, 504],  # Retry on HTTP status code 500
            allowed_methods=["GET"],  # Retry only on GET requests
            backoff_factor=1,  # Exponential backoff factor
        )



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, Field, field_validator
from typing import Any, ClassVar, Dict, List, Optional
from typing_extensions import Annotated
from typing import Optional, Set
from typing_extensions import Self

class CreateSolanaAccountRequest(BaseModel):
    """
    CreateSolanaAccountRequest
    """ # noqa: E501
    name: Optional[Annotated[str, Field(strict=True)]] = Field(default=None, description="An optional name for the account. Account names can consist of alphanumeric characters and hyphens, and be between 2 and 36 characters long. Account names must be unique across all Solana accounts in the developer's CDP Project.")
    account_policy: Optional[Annotated[str, Field(strict=True)]] = Field(default=None, description="The ID of the account-level policy to apply to the account.", alias="accountPolicy")
    __properties: ClassVar[List[str]] = ["name", "accountPolicy"]

    @field_validator('name')
    def name_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^[A-Za-z0-9][A-Za-z0-9-]{0,34}[A-Za-z0-9]$", value):
            raise ValueError(r"must validate the regular expression /^[A-Za-z0-9][A-Za-z0-9-]{0,34}[A-Za-z0-9]$/")
        return value

    @field_validator('account_policy')
    def account_policy_validate_regular_expression(cls, value):
        """Validates the regular expression"""
        if value is None:
            return value

        if not re.match(r"^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$", value):
            raise ValueError(r"must validate the regular expression /^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$/")
        return value

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )


    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of CreateSolanaAccountRequest from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([
        ])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of CreateSolanaAccountRequest from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "name": obj.get("name"),
            "accountPolicy": obj.get("accountPolicy")
        })
        return _obj





import asyncio
import contextlib
import functools
import hashlib
import inspect
import json
import os
import time
import traceback

import requests
from pydantic import BaseModel

from cdp.__version__ import __version__
from cdp.openapi_client.errors import ApiError

# This is a public client id for the analytics service
public_client_id = "54f2ee2fb3d2b901a829940d70fbfc13"


class ErrorEventData(BaseModel):
    """The data in an error event."""

    method: str  # The API method where the error occurred, e.g. createAccount, getAccount
    message: str  # The error message
    name: str  # The name of the event. This should match the name in AEC
    stack: str | None = None  # The error stack trace


EventData = ErrorEventData

Analytics = {
    "identifier": "",  # set in cdp_client.py
}


async def send_event(event: EventData) -> None:
    """Send an analytics event to the default endpoint.

    Args:
        event: The event data containing event-specific fields

    Returns:
        None - resolves when the event is sent

    """
    if os.getenv("DISABLE_CDP_ERROR_REPORTING") == "true":
        return

    timestamp = int(time.time() * 1000)

    enhanced_event = {
        "user_id": Analytics["identifier"],
        "event_type": event.name,
        "platform": "server",
        "timestamp": timestamp,
        "event_properties": {
            "project_name": "cdp-sdk",
            "cdp_sdk_language": "python",
            "version": __version__,
            **event.model_dump(),
        },
    }

    events = [enhanced_event]
    stringified_event_data = json.dumps(events)
    upload_time = str(timestamp)

    checksum = hashlib.md5((stringified_event_data + upload_time).encode("utf-8")).hexdigest()

    analytics_service_data = {
        "client": public_client_id,
        "e": stringified_event_data,
        "checksum": checksum,
    }

    api_endpoint = "https://cca-lite.coinbase.com"
    event_path = "/amp"
    event_endpoint = f"{api_endpoint}{event_path}"

    requests.post(
        event_endpoint,
        headers={"Content-Type": "application/json"},
        json=analytics_service_data,
    )


def _run_async_in_sync(coroutine):
    """Run an async coroutine in a sync context.

    Args:
        coroutine: The coroutine to run

    Returns:
        Any: The result of the coroutine, or None if it fails

    """
    try:
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # Check if loop is already running (e.g., in Jupyter notebooks)
        if loop.is_running():
            # We can't run the coroutine in an already running loop
            # This is a limitation of asyncio, so we skip analytics
            return None

        return loop.run_until_complete(coroutine)
    except Exception:
        # If anything goes wrong, silently fail to avoid breaking the SDK
        return None


def wrap_with_error_tracking(func):
    """Wrap a method with error tracking.

    Args:
        func: The function to wrap.

    Returns:
        The wrapped function.

    """
    if inspect.iscoroutinefunction(func):
        # Original function is async
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except Exception as error:
                if not should_track_error(error):
                    raise error

                event_data = ErrorEventData(
                    method=func.__name__,
                    message=str(error),
                    stack=traceback.format_exc(),
                    name="error",
                )

                with contextlib.suppress(Exception):
                    await send_event(event_data)

                raise error

        return async_wrapper
    else:
        # Original function is sync
        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as error:
                if not should_track_error(error):
                    raise error

                event_data = ErrorEventData(
                    method=func.__name__,
                    message=str(error),
                    stack=traceback.format_exc(),
                    name="error",
                )

                # Try to send analytics event from sync context
                with contextlib.suppress(Exception):
                    _run_async_in_sync(send_event(event_data))

                raise error

        return sync_wrapper


def wrap_class_with_error_tracking(cls):
    """Wrap all methods of a class with error tracking.

    Args:
        cls: The class to wrap.

    Returns:
        The class with wrapped methods.

    """
    if os.getenv("DISABLE_CDP_ERROR_REPORTING") == "true":
        return cls

    for name, method in inspect.getmembers(cls, inspect.isfunction):
        if not name.startswith("__"):
            setattr(cls, name, wrap_with_error_tracking(method))
    return cls


def should_track_error(error: Exception) -> bool:
    """Determine if an error should be tracked.

    Args:
        error: The error to check.

    Returns:
        True if the error should be tracked, False otherwise.

    """
    return isinstance(error, Exception) and not isinstance(error, ApiError)



GENESIS_HASH_MAINNET = "5eykt4UsFv8P8NJdTREpY1vzqKqZKvdpKuc147dw2N9d"
GENESIS_HASH_DEVNET = "EtWTRABZaYq6iMfeYKouRu166VU2xqa1wcaWoxPkrZBG"
GENESIS_HASH_TESTNET = "4uhcVJyU9pJkvQyS88uRDiswHXSCkY3zQawwpjk2NsNY"

USDC_MAINNET_MINT_ADDRESS = "EPjFWdd5AufqSSqeM2qN1xzybapC8G4wEGGkZwyTDt1v"
USDC_DEVNET_MINT_ADDRESS = "4zMMC9srt5Ri5X14GAgXhaHii3GnPAEERYPJgZJDncDU"



from unittest.mock import patch

from cdp.auth.utils.ws import (
    get_websocket_auth_headers,
)


@patch("cdp.auth.utils.ws.generate_jwt")
def test_get_websocket_auth_headers(mock_jwt, websocket_auth_options_factory):
    """Test getting WebSocket authentication headers."""
    # Setup
    mock_jwt.return_value = "mock.websocket.jwt.token"
    options = websocket_auth_options_factory()

    # Execute
    headers = get_websocket_auth_headers(options)

    # Verify
    assert headers["Authorization"] == "Bearer mock.websocket.jwt.token"
    assert headers["Content-Type"] == "application/json"
    assert "X-Wallet-Auth" not in headers

    # Verify JWT was called with correct parameters
    mock_jwt.assert_called_once()
    jwt_options = mock_jwt.call_args[0][0]
    assert jwt_options.api_key_id == options.api_key_id
    assert jwt_options.api_key_secret == options.api_key_secret
    assert jwt_options.request_method is None
    assert jwt_options.request_host is None
    assert jwt_options.request_path is None


@patch("cdp.auth.utils.ws.generate_jwt")
def test_websocket_headers_with_custom_expiry(mock_jwt, websocket_auth_options_factory):
    """Test WebSocket headers with custom expiration time."""
    # Setup
    mock_jwt.return_value = "mock.websocket.jwt.token"
    custom_expiry = 300
    options = websocket_auth_options_factory()
    options.expires_in = custom_expiry

    # Execute
    get_websocket_auth_headers(options)

    # Verify
    mock_jwt.assert_called_once()
    jwt_options = mock_jwt.call_args[0][0]
    assert jwt_options.expires_in == custom_expiry


@patch("cdp.auth.utils.ws.generate_jwt")
@patch("cdp.auth.utils.ws._get_correlation_data")
def test_correlation_context(mock_correlation, mock_jwt, websocket_auth_options_factory):
    """Test including correlation context in WebSocket headers."""
    # Setup
    mock_jwt.return_value = "mock.websocket.jwt.token"
    mock_correlation.return_value = "test-correlation-data"
    options = websocket_auth_options_factory()
    options.source = "custom-source"
    options.source_version = "1.0.0"

    # Execute
    headers = get_websocket_auth_headers(options)

    # Verify
    mock_correlation.assert_called_once_with(options.source, options.source_version)
    assert headers["Correlation-Context"] == "test-correlation-data"



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

import warnings
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from pydantic import Field, StrictInt, StrictStr, field_validator
from typing import Optional
from typing_extensions import Annotated
from cdp.openapi_client.models.create_policy_request import CreatePolicyRequest
from cdp.openapi_client.models.list_policies200_response import ListPolicies200Response
from cdp.openapi_client.models.policy import Policy
from cdp.openapi_client.models.update_policy_request import UpdatePolicyRequest

from cdp.openapi_client.api_client import ApiClient, RequestSerialized
from cdp.openapi_client.api_response import ApiResponse
from cdp.openapi_client.rest import RESTResponseType


class PolicyEngineApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client


    @validate_call
    async def create_policy(
        self,
        create_policy_request: CreatePolicyRequest,
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Policy:
        """Create a policy

        Create a policy that can be used to govern the behavior of accounts.

        :param create_policy_request: (required)
        :type create_policy_request: CreatePolicyRequest
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_policy_serialize(
            create_policy_request=create_policy_request,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "Policy",
            '400': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def create_policy_with_http_info(
        self,
        create_policy_request: CreatePolicyRequest,
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Policy]:
        """Create a policy

        Create a policy that can be used to govern the behavior of accounts.

        :param create_policy_request: (required)
        :type create_policy_request: CreatePolicyRequest
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_policy_serialize(
            create_policy_request=create_policy_request,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "Policy",
            '400': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def create_policy_without_preload_content(
        self,
        create_policy_request: CreatePolicyRequest,
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Create a policy

        Create a policy that can be used to govern the behavior of accounts.

        :param create_policy_request: (required)
        :type create_policy_request: CreatePolicyRequest
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_policy_serialize(
            create_policy_request=create_policy_request,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "Policy",
            '400': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _create_policy_serialize(
        self,
        create_policy_request,
        x_idempotency_key,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        if x_idempotency_key is not None:
            _header_params['X-Idempotency-Key'] = x_idempotency_key
        # process the form parameters
        # process the body parameter
        if create_policy_request is not None:
            _body_params = create_policy_request


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/v2/policy-engine/policies',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def delete_policy(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to delete.")],
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> None:
        """Delete a policy

        Delete a policy by its ID. This will have the effect of removing the policy from all accounts that are currently using it.

        :param policy_id: The ID of the policy to delete. (required)
        :type policy_id: str
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_policy_serialize(
            policy_id=policy_id,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '400': "Error",
            '404': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def delete_policy_with_http_info(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to delete.")],
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[None]:
        """Delete a policy

        Delete a policy by its ID. This will have the effect of removing the policy from all accounts that are currently using it.

        :param policy_id: The ID of the policy to delete. (required)
        :type policy_id: str
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_policy_serialize(
            policy_id=policy_id,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '400': "Error",
            '404': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def delete_policy_without_preload_content(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to delete.")],
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Delete a policy

        Delete a policy by its ID. This will have the effect of removing the policy from all accounts that are currently using it.

        :param policy_id: The ID of the policy to delete. (required)
        :type policy_id: str
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_policy_serialize(
            policy_id=policy_id,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '400': "Error",
            '404': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _delete_policy_serialize(
        self,
        policy_id,
        x_idempotency_key,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if policy_id is not None:
            _path_params['policyId'] = policy_id
        # process the query parameters
        # process the header parameters
        if x_idempotency_key is not None:
            _header_params['X-Idempotency-Key'] = x_idempotency_key
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='DELETE',
            resource_path='/v2/policy-engine/policies/{policyId}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def get_policy_by_id(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to get.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Policy:
        """Get a policy by ID

        Get a policy by its ID.

        :param policy_id: The ID of the policy to get. (required)
        :type policy_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_policy_by_id_serialize(
            policy_id=policy_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Policy",
            '404': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def get_policy_by_id_with_http_info(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to get.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Policy]:
        """Get a policy by ID

        Get a policy by its ID.

        :param policy_id: The ID of the policy to get. (required)
        :type policy_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_policy_by_id_serialize(
            policy_id=policy_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Policy",
            '404': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def get_policy_by_id_without_preload_content(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to get.")],
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get a policy by ID

        Get a policy by its ID.

        :param policy_id: The ID of the policy to get. (required)
        :type policy_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_policy_by_id_serialize(
            policy_id=policy_id,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Policy",
            '404': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _get_policy_by_id_serialize(
        self,
        policy_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if policy_id is not None:
            _path_params['policyId'] = policy_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/v2/policy-engine/policies/{policyId}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def list_policies(
        self,
        page_size: Annotated[Optional[StrictInt], Field(description="The number of policies to return per page.")] = None,
        page_token: Annotated[Optional[StrictStr], Field(description="The token for the next page of policies, if any.")] = None,
        scope: Annotated[Optional[StrictStr], Field(description="The scope of the policies to return. If `project`, the response will include exactly one policy, which is the project-level policy. If `account`, the response will include all account-level policies for the developer's CDP Project.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ListPolicies200Response:
        """List policies

        Lists the policies belonging to the developer's CDP Project. Use the `scope` parameter to filter the policies by scope. The response is paginated, and by default, returns 20 policies per page.

        :param page_size: The number of policies to return per page.
        :type page_size: int
        :param page_token: The token for the next page of policies, if any.
        :type page_token: str
        :param scope: The scope of the policies to return. If `project`, the response will include exactly one policy, which is the project-level policy. If `account`, the response will include all account-level policies for the developer's CDP Project.
        :type scope: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_policies_serialize(
            page_size=page_size,
            page_token=page_token,
            scope=scope,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ListPolicies200Response",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def list_policies_with_http_info(
        self,
        page_size: Annotated[Optional[StrictInt], Field(description="The number of policies to return per page.")] = None,
        page_token: Annotated[Optional[StrictStr], Field(description="The token for the next page of policies, if any.")] = None,
        scope: Annotated[Optional[StrictStr], Field(description="The scope of the policies to return. If `project`, the response will include exactly one policy, which is the project-level policy. If `account`, the response will include all account-level policies for the developer's CDP Project.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[ListPolicies200Response]:
        """List policies

        Lists the policies belonging to the developer's CDP Project. Use the `scope` parameter to filter the policies by scope. The response is paginated, and by default, returns 20 policies per page.

        :param page_size: The number of policies to return per page.
        :type page_size: int
        :param page_token: The token for the next page of policies, if any.
        :type page_token: str
        :param scope: The scope of the policies to return. If `project`, the response will include exactly one policy, which is the project-level policy. If `account`, the response will include all account-level policies for the developer's CDP Project.
        :type scope: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_policies_serialize(
            page_size=page_size,
            page_token=page_token,
            scope=scope,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ListPolicies200Response",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def list_policies_without_preload_content(
        self,
        page_size: Annotated[Optional[StrictInt], Field(description="The number of policies to return per page.")] = None,
        page_token: Annotated[Optional[StrictStr], Field(description="The token for the next page of policies, if any.")] = None,
        scope: Annotated[Optional[StrictStr], Field(description="The scope of the policies to return. If `project`, the response will include exactly one policy, which is the project-level policy. If `account`, the response will include all account-level policies for the developer's CDP Project.")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """List policies

        Lists the policies belonging to the developer's CDP Project. Use the `scope` parameter to filter the policies by scope. The response is paginated, and by default, returns 20 policies per page.

        :param page_size: The number of policies to return per page.
        :type page_size: int
        :param page_token: The token for the next page of policies, if any.
        :type page_token: str
        :param scope: The scope of the policies to return. If `project`, the response will include exactly one policy, which is the project-level policy. If `account`, the response will include all account-level policies for the developer's CDP Project.
        :type scope: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._list_policies_serialize(
            page_size=page_size,
            page_token=page_token,
            scope=scope,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "ListPolicies200Response",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _list_policies_serialize(
        self,
        page_size,
        page_token,
        scope,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if page_size is not None:
            
            _query_params.append(('pageSize', page_size))
            
        if page_token is not None:
            
            _query_params.append(('pageToken', page_token))
            
        if scope is not None:
            
            _query_params.append(('scope', scope))
            
        # process the header parameters
        # process the form parameters
        # process the body parameter


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )


        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/v2/policy-engine/policies',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )




    @validate_call
    async def update_policy(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to update.")],
        update_policy_request: UpdatePolicyRequest,
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> Policy:
        """Update a policy

        Updates a policy by its ID. This will have the effect of applying the updated policy to all accounts that are currently using it.

        :param policy_id: The ID of the policy to update. (required)
        :type policy_id: str
        :param update_policy_request: (required)
        :type update_policy_request: UpdatePolicyRequest
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._update_policy_serialize(
            policy_id=policy_id,
            update_policy_request=update_policy_request,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Policy",
            '400': "Error",
            '404': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data


    @validate_call
    async def update_policy_with_http_info(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to update.")],
        update_policy_request: UpdatePolicyRequest,
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[Policy]:
        """Update a policy

        Updates a policy by its ID. This will have the effect of applying the updated policy to all accounts that are currently using it.

        :param policy_id: The ID of the policy to update. (required)
        :type policy_id: str
        :param update_policy_request: (required)
        :type update_policy_request: UpdatePolicyRequest
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._update_policy_serialize(
            policy_id=policy_id,
            update_policy_request=update_policy_request,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Policy",
            '400': "Error",
            '404': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        await response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )


    @validate_call
    async def update_policy_without_preload_content(
        self,
        policy_id: Annotated[str, Field(strict=True, description="The ID of the policy to update.")],
        update_policy_request: UpdatePolicyRequest,
        x_idempotency_key: Annotated[Optional[Annotated[str, Field(min_length=36, strict=True, max_length=36)]], Field(description="An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. ")] = None,
        _request_timeout: Union[
            None,
            Annotated[StrictFloat, Field(gt=0)],
            Tuple[
                Annotated[StrictFloat, Field(gt=0)],
                Annotated[StrictFloat, Field(gt=0)]
            ]
        ] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Update a policy

        Updates a policy by its ID. This will have the effect of applying the updated policy to all accounts that are currently using it.

        :param policy_id: The ID of the policy to update. (required)
        :type policy_id: str
        :param update_policy_request: (required)
        :type update_policy_request: UpdatePolicyRequest
        :param x_idempotency_key: An optional [UUID v4](https://www.uuidgenerator.net/version4) request header for making requests safely retryable. When included, duplicate requests with the same key will return identical responses.  Refer to our [Idempotency docs](https://docs.cdp.coinbase.com/api-v2/docs/idempotency) for more information on using idempotency keys. 
        :type x_idempotency_key: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._update_policy_serialize(
            policy_id=policy_id,
            update_policy_request=update_policy_request,
            x_idempotency_key=x_idempotency_key,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index
        )

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "Policy",
            '400': "Error",
            '404': "Error",
            '409': "Error",
            '422': "Error",
            '500': "Error",
            '502': "Error",
            '503': "Error",
        }
        response_data = await self.api_client.call_api(
            *_param,
            _request_timeout=_request_timeout
        )
        return response_data.response


    def _update_policy_serialize(
        self,
        policy_id,
        update_policy_request,
        x_idempotency_key,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {
        }

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[
            str, Union[str, bytes, List[str], List[bytes], List[Tuple[str, bytes]]]
        ] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if policy_id is not None:
            _path_params['policyId'] = policy_id
        # process the query parameters
        # process the header parameters
        if x_idempotency_key is not None:
            _header_params['X-Idempotency-Key'] = x_idempotency_key
        # process the form parameters
        # process the body parameter
        if update_policy_request is not None:
            _body_params = update_policy_request


        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                [
                    'application/json'
                ]
            )

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (
                self.api_client.select_header_content_type(
                    [
                        'application/json'
                    ]
                )
            )
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = [
            'apiKeyAuth'
        ]

        return self.api_client.param_serialize(
            method='PUT',
            resource_path='/v2/policy-engine/policies/{policyId}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth
        )





import pytest

from cdp.openapi_client.models.payment_method import PaymentMethod
from cdp.openapi_client.models.payment_rail_action import PaymentRailAction


@pytest.fixture
def payment_method_model_factory():
    """Create and return a factory for Payment Method fixtures."""

    def _create_payment_method_model(
        id="123e4567-e89b-12d3-a456-426614174000",
        type="card",
        currency="usd",
        actions=None,
        limits=None,
    ):
        if actions is None:
            actions = [PaymentRailAction(PaymentRailAction.SOURCE)]
        return PaymentMethod(id=id, type=type, currency=currency, actions=actions, limits=limits)

    return _create_payment_method_model



import pytest
from eth_account import Account


@pytest.fixture
def local_account_factory():
    """Create and return a factory for test accounts."""

    def _create_local_account(private_key="0x" + "1" * 64):
        return Account.from_key(private_key)

    return _create_local_account



from typing import TypeVar

from eth_typing import HexStr

from cdp.actions.evm.transfer.types import (
    TransferExecutionStrategy,
)
from cdp.api_clients import ApiClients
from cdp.evm_server_account import EvmServerAccount
from cdp.evm_smart_account import EvmSmartAccount
from cdp.openapi_client.models.evm_user_operation import EvmUserOperation as EvmUserOperationModel

# Type for account
T = TypeVar("T", bound=EvmServerAccount | EvmSmartAccount)


async def transfer(
    api_clients: ApiClients,
    from_account: T,
    to: str | EvmServerAccount | EvmSmartAccount,
    amount: int,
    token: str,
    network: str,
    transfer_strategy: TransferExecutionStrategy,
    paymaster_url: str | None = None,
) -> HexStr | EvmUserOperationModel:
    """Transfer an amount of a token from an account to another account.

    Args:
        api_clients: The API clients to use to send the transaction
        from_account: The account to send the transaction from
        to: The account to send the transaction to
        amount: The amount of the token to transfer
        token: The token to transfer
        network: The network to transfer the token on
        transfer_strategy: The strategy to use to execute the transfer
        paymaster_url: The paymaster URL to use for the transfer. Only used for smart accounts.

    Returns:
        The result of the transfer

    """
    # Determine the recipient address
    to_address = to.address if hasattr(to, "address") else to

    kwargs = {
        "api_clients": api_clients,
        "from_account": from_account,
        "to": to_address,
        "value": amount,
        "token": token,
        "network": network,
    }

    if isinstance(from_account, EvmSmartAccount):
        kwargs["paymaster_url"] = paymaster_url

    return await transfer_strategy.execute_transfer(**kwargs)



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import unittest

from cdp.openapi_client.models.sign_evm_message_criteria_inner import SignEvmMessageCriteriaInner

class TestSignEvmMessageCriteriaInner(unittest.TestCase):
    """SignEvmMessageCriteriaInner unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional) -> SignEvmMessageCriteriaInner:
        """Test SignEvmMessageCriteriaInner
            include_optional is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # uncomment below to create an instance of `SignEvmMessageCriteriaInner`
        """
        model = SignEvmMessageCriteriaInner()
        if include_optional:
            return SignEvmMessageCriteriaInner(
                type = 'evmMessage',
                match = '^hello ([a-z]+)$'
            )
        else:
            return SignEvmMessageCriteriaInner(
                type = 'evmMessage',
                match = '^hello ([a-z]+)$',
        )
        """

    def testSignEvmMessageCriteriaInner(self):
        """Test SignEvmMessageCriteriaInner"""
        # inst_req_only = self.make_instance(include_optional=False)
        # inst_req_and_optional = self.make_instance(include_optional=True)

if __name__ == '__main__':
    unittest.main()



import json
import logging
from typing import Any
from urllib.parse import urlparse

import urllib3
from pydantic import BaseModel, Field

from cdp.auth.utils.http import GetAuthHeadersOptions, get_auth_headers

# Add logger
logger = logging.getLogger(__name__)


class Urllib3AuthClientOptions(BaseModel):
    r"""Configuration options for the authenticated HTTP client.

    Attributes:
        api_key_id - The API key ID
            Examples:
                'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'
                'organizations/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/apiKeys/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'

        api_key_secret - The API key secret
            Examples:
                'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx==' (Edwards key (Ed25519))
                '-----BEGIN EC PRIVATE KEY-----\n...\n...\n...==\n-----END EC PRIVATE KEY-----\n' (EC key (ES256))

        [wallet_secret] - Optional wallet secret for authenticating with endpoints that require wallet authentication.

        [source] - Optional source of the request

        [source_version] - Optional version of the source of the request

        [expires_in] - Optional expiration time in seconds (defaults to 120)

    """

    api_key_id: str = Field(..., description="The API key ID")
    api_key_secret: str = Field(..., description="The API key secret")
    wallet_secret: str | None = Field(None, description="Optional wallet secret")
    source: str | None = Field(None, description="Optional source identifier")
    source_version: str | None = Field(None, description="Optional source version")
    expires_in: int | None = Field(None, description="Optional JWT expiration time in seconds")


class Urllib3AuthClient:
    """HTTP client that automatically adds authentication headers."""

    def __init__(
        self,
        options: Urllib3AuthClientOptions,
        base_url: str,
        debug: bool = False,
    ):
        """Initialize the authenticated HTTP client.

        Args:
            options: The authentication configuration options
            base_url: The base URL for all requests
            debug: Whether to enable debug logging

        """
        self.options = options
        self.base_url = base_url.rstrip("/")
        self.client = urllib3.PoolManager()
        self.debug = debug

    def request(
        self,
        method: str,
        url: str,
        headers: dict[str, str] | None = None,
        body: dict[str, Any] | bytes | None = None,
        **kwargs: Any,
    ) -> urllib3.HTTPResponse:
        """Make an authenticated HTTP request.

        Args:
            method: The HTTP method
            url: The URL to request (relative or absolute)
            headers: Optional additional headers
            body: Optional request body (can be a dict for JSON or bytes)
            **kwargs: Additional arguments passed to urllib3.request()

        Returns:
            urllib3.HTTPResponse

        """
        # Handle relative URLs
        if not url.startswith("http"):
            url = f"{self.base_url}/{url.lstrip('/')}"

        # Initialize request headers and body
        request_headers = headers or {}
        body_dict = body if isinstance(body, dict) else {}
        body_bytes = json.dumps(body).encode("utf-8") if isinstance(body, dict) else body

        # Get auth headers
        parsed_url = urlparse(url)
        auth_headers = get_auth_headers(
            GetAuthHeadersOptions(
                api_key_id=self.options.api_key_id,
                api_key_secret=self.options.api_key_secret,
                request_method=method,
                request_host=parsed_url.netloc,
                request_path=parsed_url.path,
                request_body=body_dict,
                wallet_secret=self.options.wallet_secret,
                source=self.options.source,
                source_version=self.options.source_version,
                expires_in=self.options.expires_in,
            )
        )

        # Merge headers
        request_headers.update(auth_headers)

        if self.debug:
            logger.debug(
                "HTTP Request: %s %s\nHeaders: %s\nBody: %s",
                method,
                url,
                request_headers,
                body_bytes.decode("utf-8") if body_bytes else None,
            )

        response = self.client.request(
            method=method, url=url, headers=request_headers, body=body_bytes, **kwargs
        )

        if self.debug:
            logger.debug(
                "HTTP Response: %s\nHeaders: %s\nBody: %s",
                response.status,
                response.headers,
                response.data.decode("utf-8"),
            )

        return response



from pydantic import BaseModel, Field

from cdp.auth.utils.http import _get_correlation_data
from cdp.auth.utils.jwt import JwtOptions, generate_jwt


class GetWebSocketAuthHeadersOptions(BaseModel):
    """Options for generating WebSocket authentication headers.

    Attributes:
        api_key_id - The API key ID
        api_key_secret - The API key secret
        [source] - Optional source identifier
        [source_version] - Optional source version
        [expires_in] - Optional JWT expiration time in seconds
        [audience] - Optional audience claim for the JWT

    """

    api_key_id: str = Field(..., description="The API key ID")
    api_key_secret: str = Field(..., description="The API key secret")
    source: str | None = Field(None, description="Optional source identifier")
    source_version: str | None = Field(None, description="Optional source version")
    expires_in: int | None = Field(None, description="Optional JWT expiration time in seconds")
    audience: list[str] | None = Field(None, description="Optional audience claim for the JWT")


def get_websocket_auth_headers(options: GetWebSocketAuthHeadersOptions) -> dict[str, str]:
    """Get authentication headers for a WebSocket connection.

    Args:
        options: The WebSocket authentication header options

    Returns:
        Dict with authentication headers

    """
    headers = {}

    # Create JWT options with null request parameters for WebSocket
    jwt_options = JwtOptions(
        api_key_id=options.api_key_id,
        api_key_secret=options.api_key_secret,
        request_method=None,
        request_host=None,
        request_path=None,
        expires_in=options.expires_in,
        audience=options.audience,
    )

    # Generate and add JWT token
    jwt_token = generate_jwt(jwt_options)
    headers["Authorization"] = f"Bearer {jwt_token}"
    headers["Content-Type"] = "application/json"

    # Add correlation data
    headers["Correlation-Context"] = _get_correlation_data(
        options.source,
        options.source_version,
    )

    return headers



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import unittest

from cdp.openapi_client.api.payments_alpha_api import PaymentsAlphaApi


class TestPaymentsAlphaApi(unittest.IsolatedAsyncioTestCase):
    """PaymentsAlphaApi unit test stubs"""

    async def asyncSetUp(self) -> None:
        self.api = PaymentsAlphaApi()

    async def asyncTearDown(self) -> None:
        await self.api.api_client.close()

    async def test_create_payment_transfer_quote(self) -> None:
        """Test case for create_payment_transfer_quote

        Create a transfer quote
        """
        pass

    async def test_execute_payment_transfer_quote(self) -> None:
        """Test case for execute_payment_transfer_quote

        Execute a transfer quote
        """
        pass

    async def test_get_crypto_rails(self) -> None:
        """Test case for get_crypto_rails

        Get the crypto rails
        """
        pass

    async def test_get_payment_methods(self) -> None:
        """Test case for get_payment_methods

        Get the fiat payment methods
        """
        pass

    async def test_get_payment_transfer(self) -> None:
        """Test case for get_payment_transfer

        Get a transfer by ID
        """
        pass


if __name__ == '__main__':
    unittest.main()



import json
import os
from unittest.mock import MagicMock, patch

import pytest

from cdp.analytics import Analytics, ErrorEventData


@pytest.mark.asyncio
@patch("requests.post")
async def test_send_event(mock_post, mock_send_event):
    """Test sending an error event."""
    # Temporarily disable the environment variable
    original_env = os.environ.get("DISABLE_CDP_ERROR_REPORTING")
    os.environ["DISABLE_CDP_ERROR_REPORTING"] = "false"

    try:
        mock_response = MagicMock()
        mock_response.ok = True
        mock_response.status_code = 200
        mock_post.return_value = mock_response

        original_send_event = mock_send_event.original

        Analytics["identifier"] = "test-api-key-id"
        event_data = ErrorEventData(name="error", method="test", message="test")

        await original_send_event(event_data)

        mock_post.assert_called_once()

        args, kwargs = mock_post.call_args
        assert args[0] == "https://cca-lite.coinbase.com/amp"

        assert kwargs["headers"] == {"Content-Type": "application/json"}

        data = kwargs["json"]
        assert "e" in data

        event_data = json.loads(data["e"])
        assert len(event_data) > 0
        assert event_data[0]["event_type"] == "error"
        assert event_data[0]["platform"] == "server"
        assert event_data[0]["user_id"] == "test-api-key-id"
        assert event_data[0]["timestamp"] is not None

        event_props = event_data[0]["event_properties"]
        assert event_props["cdp_sdk_language"] == "python"
        assert event_props["name"] == "error"
        assert event_props["method"] == "test"
        assert event_props["message"] == "test"
    finally:
        # Restore the original environment variable
        if original_env is not None:
            os.environ["DISABLE_CDP_ERROR_REPORTING"] = original_env
        else:
            del os.environ["DISABLE_CDP_ERROR_REPORTING"]



# coding: utf-8

"""
    Coinbase Developer Platform APIs

    The Coinbase Developer Platform APIs - leading the world's transition onchain.

    The version of the OpenAPI document: 2.0.0
    Contact: cdp@coinbase.com
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


import unittest

from cdp.openapi_client.api.faucets_api import FaucetsApi


class TestFaucetsApi(unittest.IsolatedAsyncioTestCase):
    """FaucetsApi unit test stubs"""

    async def asyncSetUp(self) -> None:
        self.api = FaucetsApi()

    async def asyncTearDown(self) -> None:
        await self.api.api_client.close()

    async def test_request_evm_faucet(self) -> None:
        """Test case for request_evm_faucet

        Request funds on EVM test networks
        """
        pass

    async def test_request_solana_faucet(self) -> None:
        """Test case for request_solana_faucet

        Request funds on Solana devnet
        """
        pass


if __name__ == '__main__':
    unittest.main()
